@@ -22,19 +22,33 @@
 #endif
 
 #include <string.h>
-#include "red_common.h"
+#include <spice/vd_agent.h>
+#include "red-common.h"
 #include "agent-msg-filter.h"
-#include "red_dispatcher.h"
+#include "reds.h"
+#include "red-qxl.h"
 
-void agent_msg_filter_init(struct AgentMsgFilter *filter,
-    int copy_paste, int discard_all)
+void agent_msg_filter_config(AgentMsgFilter *filter,
+                             gboolean copy_paste, gboolean file_xfer,
+                             gboolean use_client_monitors_config)
+{
+    filter->copy_paste_enabled = copy_paste;
+    filter->file_xfer_enabled = file_xfer;
+    filter->use_client_monitors_config = use_client_monitors_config;
+}
+
+void agent_msg_filter_init(AgentMsgFilter *filter,
+                           gboolean copy_paste, gboolean file_xfer,
+                           gboolean use_client_monitors_config,
+                           int discard_all)
 {
     memset(filter, 0, sizeof(*filter));
-    filter->copy_paste_enabled = copy_paste;
+    agent_msg_filter_config(filter, copy_paste, file_xfer,
+                            use_client_monitors_config);
     filter->discard_all = discard_all;
 }
 
-int agent_msg_filter_process_data(struct AgentMsgFilter *filter,
+int agent_msg_filter_process_data(AgentMsgFilter *filter,
                                   uint8_t *data, uint32_t len)
 {
     struct VDAgentMessage msg_header;
@@ -81,8 +95,17 @@ data_to_read:
                 filter->result = AGENT_MSG_FILTER_DISCARD;
             }
             break;
+        case VD_AGENT_FILE_XFER_START:
+        case VD_AGENT_FILE_XFER_STATUS:
+        case VD_AGENT_FILE_XFER_DATA:
+            if (filter->file_xfer_enabled) {
+                filter->result = AGENT_MSG_FILTER_OK;
+            } else {
+                filter->result = AGENT_MSG_FILTER_DISCARD;
+            }
+            break;
         case VD_AGENT_MONITORS_CONFIG:
-            if (red_dispatcher_use_client_monitors_config()) {
+            if (filter->use_client_monitors_config) {
                 filter->result = AGENT_MSG_FILTER_MONITORS_CONFIG;
             } else {
                 filter->result = AGENT_MSG_FILTER_OK;
