@@ -1,6 +1,6 @@
 /* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
-   Copyright (C) 2010 Red Hat, Inc.
+   Copyright (C) 2010-2016 Red Hat, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -29,10 +29,10 @@
 #endif
 
 #include "reds.h"
-#include "char_device.h"
-#include "red_channel.h"
+#include "char-device.h"
+#include "red-channel.h"
 #include "smartcard.h"
-#include "migration_protocol.h"
+#include "migration-protocol.h"
 
 /*
  * TODO: the code doesn't really support multiple readers.
@@ -46,21 +46,25 @@
  */
 #define SMARTCARD_MAX_READERS 10
 
-typedef struct SmartCardDeviceState SmartCardDeviceState;
+// Maximal length of APDU
+#define APDUBufSize 270
 
 typedef struct SmartCardChannelClient {
     RedChannelClient base;
-    SmartCardDeviceState *smartcard_state;
+    RedCharDeviceSmartcard *smartcard;
 
     /* read_from_client/write_to_device buffer.
      * The beginning of the buffer should always be VSCMsgHeader*/
-    SpiceCharDeviceWriteBuffer *write_buf;
-    int msg_in_write_buf; /* was the client msg received into a SpiceCharDeviceWriteBuffer
+    RedCharDeviceWriteBuffer *write_buf;
+    int msg_in_write_buf; /* was the client msg received into a RedCharDeviceWriteBuffer
                            * or was it explicitly malloced */
 } SmartCardChannelClient;
 
-struct SmartCardDeviceState {
-    SpiceCharDeviceState *chardev_st;
+G_DEFINE_TYPE(RedCharDeviceSmartcard, red_char_device_smartcard, RED_TYPE_CHAR_DEVICE)
+
+#define RED_CHAR_DEVICE_SMARTCARD_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), RED_TYPE_CHAR_DEVICE_SMARTCARD, RedCharDeviceSmartcardPrivate))
+
+struct RedCharDeviceSmartcardPrivate {
     uint32_t             reader_id;
     /* read_from_device buffer */
     uint8_t             *buf;
@@ -73,28 +77,25 @@ struct SmartCardDeviceState {
 };
 
 enum {
-    PIPE_ITEM_TYPE_ERROR = PIPE_ITEM_TYPE_CHANNEL_BASE,
-    PIPE_ITEM_TYPE_SMARTCARD_DATA,
-    PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA,
+    RED_PIPE_ITEM_TYPE_ERROR = RED_PIPE_ITEM_TYPE_CHANNEL_BASE,
+    RED_PIPE_ITEM_TYPE_SMARTCARD_DATA,
+    RED_PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA,
 };
 
-typedef struct ErrorItem {
-    PipeItem base;
+typedef struct RedErrorItem {
+    RedPipeItem base;
     VSCMsgHeader vheader;
     VSCMsgError  error;
-} ErrorItem;
+} RedErrorItem;
 
-typedef struct MsgItem {
-    PipeItem base;
-    uint32_t refs;
+typedef struct RedMsgItem {
+    RedPipeItem base;
 
     VSCMsgHeader* vheader;
-} MsgItem;
+} RedMsgItem;
 
-static MsgItem *smartcard_get_vsc_msg_item(RedChannelClient *rcc, VSCMsgHeader *vheader);
-static MsgItem *smartcard_ref_vsc_msg_item(MsgItem *item);
-static void smartcard_unref_vsc_msg_item(MsgItem *item);
-static void smartcard_channel_client_pipe_add_push(RedChannelClient *rcc, PipeItem *item);
+static RedMsgItem *smartcard_get_vsc_msg_item(RedChannelClient *rcc, VSCMsgHeader *vheader);
+static void smartcard_channel_client_pipe_add_push(RedChannelClient *rcc, RedPipeItem *item);
 
 typedef struct SmartCardChannel {
     RedChannel base;
@@ -107,85 +108,73 @@ static struct Readers {
 
 static SpiceCharDeviceInstance* smartcard_readers_get_unattached(void);
 static SpiceCharDeviceInstance* smartcard_readers_get(uint32_t reader_id);
-static int smartcard_char_device_add_to_readers(SpiceCharDeviceInstance *sin);
+static int smartcard_char_device_add_to_readers(RedsState *reds, SpiceCharDeviceInstance *sin);
 static void smartcard_char_device_attach_client(
     SpiceCharDeviceInstance *char_device, SmartCardChannelClient *scc);
-static void smartcard_channel_write_to_reader(SpiceCharDeviceWriteBuffer *write_buf);
+static void smartcard_channel_write_to_reader(RedCharDeviceWriteBuffer *write_buf);
 
-static MsgItem *smartcard_char_device_on_message_from_device(
-    SmartCardDeviceState *state, VSCMsgHeader *header);
-static SmartCardDeviceState *smartcard_device_state_new(SpiceCharDeviceInstance *sin);
-static void smartcard_device_state_free(SmartCardDeviceState* st);
-static void smartcard_init(void);
+static RedMsgItem *smartcard_char_device_on_message_from_device(
+    RedCharDeviceSmartcard *dev, VSCMsgHeader *header);
+static RedCharDeviceSmartcard *smartcard_device_new(RedsState *reds, SpiceCharDeviceInstance *sin);
+static void smartcard_init(RedsState *reds);
 
-static void smartcard_read_buf_prepare(SmartCardDeviceState *state, VSCMsgHeader *vheader)
+static void smartcard_read_buf_prepare(RedCharDeviceSmartcard *dev, VSCMsgHeader *vheader)
 {
     uint32_t msg_len;
 
     msg_len = ntohl(vheader->length);
-    if (msg_len > state->buf_size) {
-        state->buf_size = MAX(state->buf_size * 2, msg_len + sizeof(VSCMsgHeader));
-        state->buf = spice_realloc(state->buf, state->buf_size);
+    if (msg_len > dev->priv->buf_size) {
+        dev->priv->buf_size = MAX(dev->priv->buf_size * 2, msg_len + sizeof(VSCMsgHeader));
+        dev->priv->buf = spice_realloc(dev->priv->buf, dev->priv->buf_size);
     }
 }
 
-SpiceCharDeviceMsgToClient *smartcard_read_msg_from_device(SpiceCharDeviceInstance *sin,
-                                                           void *opaque)
+static RedPipeItem *smartcard_read_msg_from_device(SpiceCharDeviceInstance *sin,
+                                                   void *opaque)
 {
-    SmartCardDeviceState *state = opaque;
-    SpiceCharDeviceInterface *sif = SPICE_CONTAINEROF(sin->base.sif, SpiceCharDeviceInterface, base);
-    VSCMsgHeader *vheader = (VSCMsgHeader*)state->buf;
+    RedCharDeviceSmartcard *dev = opaque;
+    SpiceCharDeviceInterface *sif = spice_char_device_get_interface(sin);
+    VSCMsgHeader *vheader = (VSCMsgHeader*)dev->priv->buf;
     int n;
     int remaining;
     int actual_length;
 
-    while ((n = sif->read(sin, state->buf_pos, state->buf_size - state->buf_used)) > 0) {
-        MsgItem *msg_to_client;
+    while ((n = sif->read(sin, dev->priv->buf_pos, dev->priv->buf_size - dev->priv->buf_used)) > 0) {
+        RedMsgItem *msg_to_client;
 
-        state->buf_pos += n;
-        state->buf_used += n;
-        if (state->buf_used < sizeof(VSCMsgHeader)) {
+        dev->priv->buf_pos += n;
+        dev->priv->buf_used += n;
+        if (dev->priv->buf_used < sizeof(VSCMsgHeader)) {
             continue;
         }
-        smartcard_read_buf_prepare(state, vheader);
+        smartcard_read_buf_prepare(dev, vheader);
         actual_length = ntohl(vheader->length);
-        if (state->buf_used - sizeof(VSCMsgHeader) < actual_length) {
+        if (dev->priv->buf_used - sizeof(VSCMsgHeader) < actual_length) {
             continue;
         }
-        msg_to_client = smartcard_char_device_on_message_from_device(state, vheader);
-        remaining = state->buf_used - sizeof(VSCMsgHeader) - actual_length;
+        msg_to_client = smartcard_char_device_on_message_from_device(dev, vheader);
+        remaining = dev->priv->buf_used - sizeof(VSCMsgHeader) - actual_length;
         if (remaining > 0) {
-            memcpy(state->buf, state->buf_pos, remaining);
+            memcpy(dev->priv->buf, dev->priv->buf_pos, remaining);
         }
-        state->buf_pos = state->buf;
-        state->buf_used = remaining;
+        dev->priv->buf_pos = dev->priv->buf;
+        dev->priv->buf_used = remaining;
         if (msg_to_client) {
-            return msg_to_client;
+            return &msg_to_client->base;
         }
     }
     return NULL;
 }
 
-static SpiceCharDeviceMsgToClient *smartcard_ref_msg_to_client(SpiceCharDeviceMsgToClient *msg,
-                                                               void *opaque)
-{
-    return smartcard_ref_vsc_msg_item((MsgItem *)msg);
-}
-
-static void smartcard_unref_msg_to_client(SpiceCharDeviceMsgToClient *msg,
-                                          void *opaque)
-{
-    smartcard_unref_vsc_msg_item((MsgItem *)msg);
-}
-
-static void smartcard_send_msg_to_client(SpiceCharDeviceMsgToClient *msg,
+static void smartcard_send_msg_to_client(RedPipeItem *msg,
                                          RedClient *client,
                                          void *opaque)
 {
-    SmartCardDeviceState *dev = opaque;
-    spice_assert(dev->scc && dev->scc->base.client == client);
-    smartcard_channel_client_pipe_add_push(&dev->scc->base, &((MsgItem *)msg)->base);
+    RedCharDeviceSmartcard *dev = opaque;
 
+    spice_assert(dev->priv->scc && dev->priv->scc->base.client == client);
+    red_pipe_item_ref(msg);
+    smartcard_channel_client_pipe_add_push(&dev->priv->scc->base, msg);
 }
 
 static void smartcard_send_tokens_to_client(RedClient *client, uint32_t tokens, void *opaque)
@@ -195,15 +184,15 @@ static void smartcard_send_tokens_to_client(RedClient *client, uint32_t tokens,
 
 static void smartcard_remove_client(RedClient *client, void *opaque)
 {
-    SmartCardDeviceState *dev = opaque;
+    RedCharDeviceSmartcard *dev = opaque;
 
-    spice_printerr("smartcard  state %p, client %p", dev, client);
-    spice_assert(dev->scc && dev->scc->base.client == client);
-    red_channel_client_shutdown(&dev->scc->base);
+    spice_printerr("smartcard  dev %p, client %p", dev, client);
+    spice_assert(dev->priv->scc && dev->priv->scc->base.client == client);
+    red_channel_client_shutdown(&dev->priv->scc->base);
 }
 
-MsgItem *smartcard_char_device_on_message_from_device(SmartCardDeviceState *state,
-                                                      VSCMsgHeader *vheader)
+RedMsgItem *smartcard_char_device_on_message_from_device(RedCharDeviceSmartcard *dev,
+                                                         VSCMsgHeader *vheader)
 {
     VSCMsgHeader *sent_header;
 
@@ -218,29 +207,29 @@ MsgItem *smartcard_char_device_on_message_from_device(SmartCardDeviceState *stat
             break;
     }
     /* We pass any VSC_Error right now - might need to ignore some? */
-    if (state->reader_id == VSCARD_UNDEFINED_READER_ID && vheader->type != VSC_Init) {
+    if (dev->priv->reader_id == VSCARD_UNDEFINED_READER_ID && vheader->type != VSC_Init) {
         spice_printerr("error: reader_id not assigned for message of type %d", vheader->type);
     }
-    if (state->scc) {
+    if (dev->priv->scc) {
         sent_header = spice_memdup(vheader, sizeof(*vheader) + vheader->length);
         /* We patch the reader_id, since the device only knows about itself, and
          * we know about the sum of readers. */
-        sent_header->reader_id = state->reader_id;
-        return smartcard_get_vsc_msg_item(&state->scc->base, sent_header);
+        sent_header->reader_id = dev->priv->reader_id;
+        return smartcard_get_vsc_msg_item(&dev->priv->scc->base, sent_header);
     }
     return NULL;
 }
 
-static int smartcard_char_device_add_to_readers(SpiceCharDeviceInstance *char_device)
+static int smartcard_char_device_add_to_readers(RedsState *reds, SpiceCharDeviceInstance *char_device)
 {
-    SmartCardDeviceState *state = spice_char_device_state_opaque_get(char_device->st);
+    RedCharDeviceSmartcard *dev = red_char_device_opaque_get(char_device->st);
 
     if (g_smartcard_readers.num >= SMARTCARD_MAX_READERS) {
         return -1;
     }
-    state->reader_id = g_smartcard_readers.num;
+    dev->priv->reader_id = g_smartcard_readers.num;
     g_smartcard_readers.sin[g_smartcard_readers.num++] = char_device;
-    smartcard_init();
+    smartcard_init(reds);
     return 0;
 }
 
@@ -255,88 +244,66 @@ static SpiceCharDeviceInstance *smartcard_readers_get(uint32_t reader_id)
 static SpiceCharDeviceInstance *smartcard_readers_get_unattached(void)
 {
     int i;
-    SmartCardDeviceState* state;
+    RedCharDeviceSmartcard* dev;
 
     for (i = 0; i < g_smartcard_readers.num; ++i) {
-        state = spice_char_device_state_opaque_get(g_smartcard_readers.sin[i]->st);
-        if (!state->scc) {
+        dev = red_char_device_opaque_get(g_smartcard_readers.sin[i]->st);
+        if (!dev->priv->scc) {
             return g_smartcard_readers.sin[i];
         }
     }
     return NULL;
 }
 
-static SmartCardDeviceState *smartcard_device_state_new(SpiceCharDeviceInstance *sin)
+static RedCharDeviceSmartcard *smartcard_device_new(RedsState *reds, SpiceCharDeviceInstance *sin)
 {
-    SmartCardDeviceState *st;
-    SpiceCharDeviceCallbacks chardev_cbs = { NULL, };
+    RedCharDevice *char_dev;
 
-    chardev_cbs.read_one_msg_from_device = smartcard_read_msg_from_device;
-    chardev_cbs.ref_msg_to_client = smartcard_ref_msg_to_client;
-    chardev_cbs.unref_msg_to_client = smartcard_unref_msg_to_client;
-    chardev_cbs.send_msg_to_client = smartcard_send_msg_to_client;
-    chardev_cbs.send_tokens_to_client = smartcard_send_tokens_to_client;
-    chardev_cbs.remove_client = smartcard_remove_client;
+    char_dev = g_object_new(RED_TYPE_CHAR_DEVICE_SMARTCARD,
+                            "sin", sin,
+                            "spice-server", reds,
+                            "client-tokens-interval", 0ULL,
+                            "self-tokens", ~0ULL,
+                            NULL);
 
-    st = spice_new0(SmartCardDeviceState, 1);
-    st->chardev_st = spice_char_device_state_create(sin,
-                                                    0, /* tokens interval */
-                                                    ~0, /* self tokens */
-                                                    &chardev_cbs,
-                                                    st);
-    st->reader_id = VSCARD_UNDEFINED_READER_ID;
-    st->reader_added = FALSE;
-    st->buf_size = APDUBufSize + sizeof(VSCMsgHeader);
-    st->buf = spice_malloc(st->buf_size);
-    st->buf_pos = st->buf;
-    st->buf_used = 0;
-    st->scc = NULL;
-    return st;
-}
+    g_object_set(char_dev, "opaque", char_dev, NULL);
 
-static void smartcard_device_state_free(SmartCardDeviceState* st)
-{
-    if (st->scc) {
-        st->scc->smartcard_state = NULL;
-    }
-    free(st->buf);
-    spice_char_device_state_destroy(st->chardev_st);
-    free(st);
+    return RED_CHAR_DEVICE_SMARTCARD(char_dev);
 }
 
 void smartcard_device_disconnect(SpiceCharDeviceInstance *char_device)
 {
-    SmartCardDeviceState *st = spice_char_device_state_opaque_get(char_device->st);
+    g_return_if_fail(RED_IS_CHAR_DEVICE_SMARTCARD(char_device->st));
 
-    smartcard_device_state_free(st);
+    g_object_unref(char_device->st);
 }
 
-SpiceCharDeviceState *smartcard_device_connect(SpiceCharDeviceInstance *char_device)
+RedCharDevice *smartcard_device_connect(RedsState *reds, SpiceCharDeviceInstance *char_device)
 {
-    SmartCardDeviceState *st;
+    RedCharDeviceSmartcard *dev;
 
-    st = smartcard_device_state_new(char_device);
-    if (smartcard_char_device_add_to_readers(char_device) == -1) {
-        smartcard_device_state_free(st);
+    dev = smartcard_device_new(reds, char_device);
+    if (smartcard_char_device_add_to_readers(reds, char_device) == -1) {
+        g_object_unref(dev);
         return NULL;
     }
-    return st->chardev_st;
+    return RED_CHAR_DEVICE(dev);
 }
 
-static void smartcard_char_device_notify_reader_add(SmartCardDeviceState *st)
+static void smartcard_char_device_notify_reader_add(RedCharDeviceSmartcard *dev)
 {
-    SpiceCharDeviceWriteBuffer *write_buf;
+    RedCharDeviceWriteBuffer *write_buf;
     VSCMsgHeader *vheader;
 
-    write_buf = spice_char_device_write_buffer_get(st->chardev_st, NULL, sizeof(vheader));
+    write_buf = red_char_device_write_buffer_get(RED_CHAR_DEVICE(dev), NULL, sizeof(*vheader));
     if (!write_buf) {
         spice_error("failed to allocate write buffer");
         return;
     }
-    st->reader_added = TRUE;
+    dev->priv->reader_added = TRUE;
     vheader = (VSCMsgHeader *)write_buf->buf;
     vheader->type = VSC_ReaderAdd;
-    vheader->reader_id = st->reader_id;
+    vheader->reader_id = dev->priv->reader_id;
     vheader->length = 0;
     smartcard_channel_write_to_reader(write_buf);
 }
@@ -344,62 +311,62 @@ static void smartcard_char_device_notify_reader_add(SmartCardDeviceState *st)
 static void smartcard_char_device_attach_client(SpiceCharDeviceInstance *char_device,
                                                 SmartCardChannelClient *scc)
 {
-    SmartCardDeviceState *st = spice_char_device_state_opaque_get(char_device->st);
+    RedCharDeviceSmartcard *dev = red_char_device_opaque_get(char_device->st);
     int client_added;
 
-    spice_assert(!scc->smartcard_state && !st->scc);
-    st->scc = scc;
-    scc->smartcard_state = st;
-    client_added = spice_char_device_client_add(st->chardev_st,
-                                                scc->base.client,
-                                                FALSE, /* no flow control yet */
-                                                0, /* send queue size */
-                                                ~0,
-                                                ~0,
-                                                red_channel_client_waits_for_migrate_data(
-                                                    &scc->base));
+    spice_assert(!scc->smartcard && !dev->priv->scc);
+    dev->priv->scc = scc;
+    scc->smartcard = dev;
+    client_added = red_char_device_client_add(RED_CHAR_DEVICE(dev),
+                                              scc->base.client,
+                                              FALSE, /* no flow control yet */
+                                              0, /* send queue size */
+                                              ~0,
+                                              ~0,
+                                              red_channel_client_is_waiting_for_migrate_data(
+                                                  &scc->base));
     if (!client_added) {
         spice_warning("failed");
-        st->scc = NULL;
-        scc->smartcard_state = NULL;
+        dev->priv->scc = NULL;
+        scc->smartcard = NULL;
         red_channel_client_disconnect(&scc->base);
     }
 }
 
-static void smartcard_char_device_notify_reader_remove(SmartCardDeviceState *st)
+static void smartcard_char_device_notify_reader_remove(RedCharDeviceSmartcard *dev)
 {
-    SpiceCharDeviceWriteBuffer *write_buf;
+    RedCharDeviceWriteBuffer *write_buf;
     VSCMsgHeader *vheader;
 
-    if (!st->reader_added) {
+    if (!dev->priv->reader_added) {
         spice_debug("reader add was never sent to the device");
         return;
     }
-    write_buf = spice_char_device_write_buffer_get(st->chardev_st, NULL, sizeof(vheader));
+    write_buf = red_char_device_write_buffer_get(RED_CHAR_DEVICE(dev), NULL, sizeof(*vheader));
     if (!write_buf) {
         spice_error("failed to allocate write buffer");
         return;
     }
-    st->reader_added = FALSE;
+    dev->priv->reader_added = FALSE;
     vheader = (VSCMsgHeader *)write_buf->buf;
     vheader->type = VSC_ReaderRemove;
-    vheader->reader_id = st->reader_id;
+    vheader->reader_id = dev->priv->reader_id;
     vheader->length = 0;
     smartcard_channel_write_to_reader(write_buf);
 }
 
 static void smartcard_char_device_detach_client(SmartCardChannelClient *scc)
 {
-    SmartCardDeviceState *st;
+    RedCharDeviceSmartcard *dev;
 
-    if (!scc->smartcard_state) {
+    if (!scc->smartcard) {
         return;
     }
-    st = scc->smartcard_state;
-    spice_assert(st->scc == scc);
-    spice_char_device_client_remove(st->chardev_st, scc->base.client);
-    scc->smartcard_state = NULL;
-    st->scc = NULL;
+    dev = scc->smartcard;
+    spice_assert(dev->priv->scc == scc);
+    red_char_device_client_remove(RED_CHAR_DEVICE(dev), scc->base.client);
+    scc->smartcard = NULL;
+    dev->priv->scc = NULL;
 }
 
 static int smartcard_channel_client_config_socket(RedChannelClient *rcc)
@@ -416,17 +383,17 @@ static uint8_t *smartcard_channel_alloc_msg_rcv_buf(RedChannelClient *rcc,
     /* todo: only one reader is actually supported. When we fix the code to support
      * multiple readers, we will porbably associate different devices to
      * differenc channels */
-    if (!scc->smartcard_state) {
+    if (!scc->smartcard) {
         scc->msg_in_write_buf = FALSE;
         return spice_malloc(size);
     } else {
-        SmartCardDeviceState *st;
+        RedCharDeviceSmartcard *dev;
 
         spice_assert(g_smartcard_readers.num == 1);
-        st = scc->smartcard_state;
-        spice_assert(st->scc || scc->smartcard_state);
+        dev = scc->smartcard;
+        spice_assert(dev->priv->scc || scc->smartcard);
         spice_assert(!scc->write_buf);
-        scc->write_buf = spice_char_device_write_buffer_get(st->chardev_st, rcc->client, size);
+        scc->write_buf = red_char_device_write_buffer_get(RED_CHAR_DEVICE(dev), rcc->client, size);
 
         if (!scc->write_buf) {
             spice_error("failed to allocate write buffer");
@@ -452,18 +419,15 @@ static void smartcard_channel_release_msg_rcv_buf(RedChannelClient *rcc,
         spice_assert(!scc->write_buf);
         free(msg);
     } else {
-        SpiceCharDeviceState *dev_st;
         if (scc->write_buf) { /* msg hasn't been pushed to the guest */
             spice_assert(scc->write_buf->buf == msg);
-            dev_st = scc->smartcard_state ? scc->smartcard_state->chardev_st : NULL;
-            spice_char_device_write_buffer_release(dev_st, scc->write_buf);
-            scc->write_buf = NULL;
+            red_char_device_write_buffer_release(RED_CHAR_DEVICE(scc->smartcard), &scc->write_buf);
         }
     }
 }
 
 static void smartcard_channel_send_data(RedChannelClient *rcc, SpiceMarshaller *m,
-                                        PipeItem *item, VSCMsgHeader *vheader)
+                                        RedPipeItem *item, VSCMsgHeader *vheader)
 {
     spice_assert(rcc);
     spice_assert(vheader);
@@ -475,110 +439,98 @@ static void smartcard_channel_send_data(RedChannelClient *rcc, SpiceMarshaller *
 }
 
 static void smartcard_channel_send_error(
-    RedChannelClient *rcc, SpiceMarshaller *m, PipeItem *item)
+    RedChannelClient *rcc, SpiceMarshaller *m, RedPipeItem *item)
 {
-    ErrorItem* error_item = (ErrorItem*)item;
+    RedErrorItem* error_item = SPICE_UPCAST(RedErrorItem, item);
 
     smartcard_channel_send_data(rcc, m, item, &error_item->vheader);
 }
 
 static void smartcard_channel_send_msg(RedChannelClient *rcc,
-                                       SpiceMarshaller *m, PipeItem *item)
+                                       SpiceMarshaller *m, RedPipeItem *item)
 {
-    MsgItem* msg_item = (MsgItem*)item;
+    RedMsgItem* msg_item = SPICE_UPCAST(RedMsgItem, item);
 
     smartcard_channel_send_data(rcc, m, item, msg_item->vheader);
 }
 
 static void smartcard_channel_send_migrate_data(RedChannelClient *rcc,
-                                                SpiceMarshaller *m, PipeItem *item)
+                                                SpiceMarshaller *m, RedPipeItem *item)
 {
     SmartCardChannelClient *scc;
-    SmartCardDeviceState *state;
+    RedCharDeviceSmartcard *dev;
     SpiceMarshaller *m2;
 
     scc = SPICE_CONTAINEROF(rcc, SmartCardChannelClient, base);
-    state = scc->smartcard_state;
+    dev = scc->smartcard;
     red_channel_client_init_send_data(rcc, SPICE_MSG_MIGRATE_DATA, item);
     spice_marshaller_add_uint32(m, SPICE_MIGRATE_DATA_SMARTCARD_MAGIC);
     spice_marshaller_add_uint32(m, SPICE_MIGRATE_DATA_SMARTCARD_VERSION);
 
-    if (!state) {
-        spice_char_device_state_migrate_data_marshall_empty(m);
+    if (!dev) {
+        red_char_device_migrate_data_marshall_empty(m);
         spice_marshaller_add_uint8(m, 0);
         spice_marshaller_add_uint32(m, 0);
         spice_marshaller_add_uint32(m, 0);
-        spice_debug("null char dev state");
+        spice_debug("null char dev");
     } else {
-        spice_char_device_state_migrate_data_marshall(state->chardev_st, m);
-        spice_marshaller_add_uint8(m, state->reader_added);
-        spice_marshaller_add_uint32(m, state->buf_used);
+        red_char_device_migrate_data_marshall(RED_CHAR_DEVICE(dev), m);
+        spice_marshaller_add_uint8(m, dev->priv->reader_added);
+        spice_marshaller_add_uint32(m, dev->priv->buf_used);
         m2 = spice_marshaller_get_ptr_submarshaller(m, 0);
-        spice_marshaller_add(m2, state->buf, state->buf_used);
-        spice_debug("reader added %d partial read size %u", state->reader_added, state->buf_used);
+        spice_marshaller_add(m2, dev->priv->buf, dev->priv->buf_used);
+        spice_debug("reader added %d partial read size %u", dev->priv->reader_added, dev->priv->buf_used);
     }
 }
 
-static void smartcard_channel_send_item(RedChannelClient *rcc, PipeItem *item)
+static void smartcard_channel_send_item(RedChannelClient *rcc, RedPipeItem *item)
 {
     SpiceMarshaller *m = red_channel_client_get_marshaller(rcc);
 
     switch (item->type) {
-    case PIPE_ITEM_TYPE_ERROR:
+    case RED_PIPE_ITEM_TYPE_ERROR:
         smartcard_channel_send_error(rcc, m, item);
         break;
-    case PIPE_ITEM_TYPE_SMARTCARD_DATA:
+    case RED_PIPE_ITEM_TYPE_SMARTCARD_DATA:
         smartcard_channel_send_msg(rcc, m, item);
         break;
-    case PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA:
+    case RED_PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA:
         smartcard_channel_send_migrate_data(rcc, m, item);
         break;
     default:
         spice_error("bad pipe item %d", item->type);
-        free(item);
         return;
     }
     red_channel_client_begin_send_message(rcc);
 }
 
-static void smartcard_channel_release_pipe_item(RedChannelClient *rcc,
-                                      PipeItem *item, int item_pushed)
-{
-    if (item->type == PIPE_ITEM_TYPE_SMARTCARD_DATA) {
-        smartcard_unref_vsc_msg_item((MsgItem *)item);
-    } else {
-        free(item);
-    }
-}
-
 static void smartcard_channel_on_disconnect(RedChannelClient *rcc)
 {
     SmartCardChannelClient *scc = SPICE_CONTAINEROF(rcc, SmartCardChannelClient, base);
 
-    if (scc->smartcard_state) {
-        SmartCardDeviceState *st = scc->smartcard_state;
+    if (scc->smartcard) {
+        RedCharDeviceSmartcard *dev = scc->smartcard;
 
         smartcard_char_device_detach_client(scc);
-        smartcard_char_device_notify_reader_remove(st);
+        smartcard_char_device_notify_reader_remove(dev);
     }
 }
 
 /* this is called from both device input and client input. since the device is
  * a usb device, the context is still the main thread (kvm_main_loop, timers)
  * so no mutex is required. */
-static void smartcard_channel_client_pipe_add_push(RedChannelClient *rcc, PipeItem *item)
+static void smartcard_channel_client_pipe_add_push(RedChannelClient *rcc,
+                                                   RedPipeItem *item)
 {
     red_channel_client_pipe_add_push(rcc, item);
 }
 
 static void smartcard_push_error(RedChannelClient *rcc, uint32_t reader_id, VSCErrorCode error)
 {
-    ErrorItem *error_item = spice_new0(ErrorItem, 1);
+    RedErrorItem *error_item = spice_new0(RedErrorItem, 1);
 
-    red_channel_pipe_item_init(rcc->channel, &error_item->base,
-                               PIPE_ITEM_TYPE_ERROR);
+    red_pipe_item_init(&error_item->base, RED_PIPE_ITEM_TYPE_ERROR);
 
-    error_item->base.type = PIPE_ITEM_TYPE_ERROR;
     error_item->vheader.reader_id = reader_id;
     error_item->vheader.type = VSC_Error;
     error_item->vheader.length = sizeof(error_item->error);
@@ -586,35 +538,28 @@ static void smartcard_push_error(RedChannelClient *rcc, uint32_t reader_id, VSCE
     smartcard_channel_client_pipe_add_push(rcc, &error_item->base);
 }
 
-static MsgItem *smartcard_get_vsc_msg_item(RedChannelClient *rcc, VSCMsgHeader *vheader)
+static void smartcard_free_vsc_msg_item(RedPipeItem *base)
 {
-    MsgItem *msg_item = spice_new0(MsgItem, 1);
+    RedMsgItem *item = SPICE_UPCAST(RedMsgItem, base);
+    free(item->vheader);
+    free(item);
+}
 
-    red_channel_pipe_item_init(rcc->channel, &msg_item->base,
-                               PIPE_ITEM_TYPE_SMARTCARD_DATA);
-    msg_item->refs = 1;
+static RedMsgItem *smartcard_get_vsc_msg_item(RedChannelClient *rcc,
+                                              VSCMsgHeader *vheader)
+{
+    RedMsgItem *msg_item = spice_new0(RedMsgItem, 1);
+
+    red_pipe_item_init_full(&msg_item->base, RED_PIPE_ITEM_TYPE_SMARTCARD_DATA,
+                            smartcard_free_vsc_msg_item);
     msg_item->vheader = vheader;
     return msg_item;
 }
 
-static MsgItem *smartcard_ref_vsc_msg_item(MsgItem *item)
-{
-    item->refs++;
-    return item;
-}
-
-static void smartcard_unref_vsc_msg_item(MsgItem *item)
-{
-    if (!--item->refs) {
-        free(item->vheader);
-        free(item);
-    }
-}
-
 static void smartcard_remove_reader(SmartCardChannelClient *scc, uint32_t reader_id)
 {
     SpiceCharDeviceInstance *char_device = smartcard_readers_get(reader_id);
-    SmartCardDeviceState *state;
+    RedCharDeviceSmartcard *dev;
 
     if (char_device == NULL) {
         smartcard_push_error(&scc->base, reader_id,
@@ -622,19 +567,19 @@ static void smartcard_remove_reader(SmartCardChannelClient *scc, uint32_t reader
         return;
     }
 
-    state = spice_char_device_state_opaque_get(char_device->st);
-    if (state->reader_added == FALSE) {
+    dev = red_char_device_opaque_get(char_device->st);
+    if (dev->priv->reader_added == FALSE) {
         smartcard_push_error(&scc->base, reader_id,
             VSC_GENERAL_ERROR);
         return;
     }
-    spice_assert(scc->smartcard_state == state);
-    smartcard_char_device_notify_reader_remove(state);
+    spice_assert(scc->smartcard == dev);
+    smartcard_char_device_notify_reader_remove(dev);
 }
 
 static void smartcard_add_reader(SmartCardChannelClient *scc, uint8_t *name)
 {
-    if (!scc->smartcard_state) { /* we already tried to attach a reader to the client
+    if (!scc->smartcard) { /* we already tried to attach a reader to the client
                                     when it connected */
         SpiceCharDeviceInstance *char_device = smartcard_readers_get_unattached();
 
@@ -645,16 +590,16 @@ static void smartcard_add_reader(SmartCardChannelClient *scc, uint8_t *name)
         }
         smartcard_char_device_attach_client(char_device, scc);
     }
-    smartcard_char_device_notify_reader_add(scc->smartcard_state);
+    smartcard_char_device_notify_reader_add(scc->smartcard);
     // The device sends a VSC_Error message, we will let it through, no
     // need to send our own. We already set the correct reader_id, from
-    // our SmartCardDeviceState.
+    // our RedCharDeviceSmartcard.
 }
 
-static void smartcard_channel_write_to_reader(SpiceCharDeviceWriteBuffer *write_buf)
+static void smartcard_channel_write_to_reader(RedCharDeviceWriteBuffer *write_buf)
 {
     SpiceCharDeviceInstance *sin;
-    SmartCardDeviceState *st;
+    RedCharDeviceSmartcard *dev;
     VSCMsgHeader *vheader;
     uint32_t actual_length;
 
@@ -663,8 +608,8 @@ static void smartcard_channel_write_to_reader(SpiceCharDeviceWriteBuffer *write_
 
     spice_assert(vheader->reader_id <= g_smartcard_readers.num);
     sin = g_smartcard_readers.sin[vheader->reader_id];
-    st = (SmartCardDeviceState *)spice_char_device_state_opaque_get(sin->st);
-    spice_assert(!st->scc || st == st->scc->smartcard_state);
+    dev = (RedCharDeviceSmartcard *)red_char_device_opaque_get(sin->st);
+    spice_assert(!dev->priv->scc || dev == dev->priv->scc->smartcard);
     /* protocol requires messages to be in network endianess */
     vheader->type = htonl(vheader->type);
     vheader->length = htonl(vheader->length);
@@ -672,33 +617,33 @@ static void smartcard_channel_write_to_reader(SpiceCharDeviceWriteBuffer *write_
     write_buf->buf_used = actual_length + sizeof(VSCMsgHeader);
     /* pushing the buffer to the write queue; It will be released
      * when it will be fully consumed by the device */
-    spice_char_device_write_buffer_add(sin->st, write_buf);
-    if (st->scc && write_buf == st->scc->write_buf) {
-        st->scc->write_buf = NULL;
+    red_char_device_write_buffer_add(sin->st, write_buf);
+    if (dev->priv->scc && write_buf == dev->priv->scc->write_buf) {
+        dev->priv->scc->write_buf = NULL;
     }
 }
 
 static int smartcard_channel_client_handle_migrate_flush_mark(RedChannelClient *rcc)
 {
-    red_channel_client_pipe_add_type(rcc, PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA);
+    red_channel_client_pipe_add_type(rcc, RED_PIPE_ITEM_TYPE_SMARTCARD_MIGRATE_DATA);
     return TRUE;
 }
 
-static void smartcard_device_state_restore_partial_read(SmartCardDeviceState *state,
-                                                        SpiceMigrateDataSmartcard *mig_data)
+static void smartcard_device_restore_partial_read(RedCharDeviceSmartcard *dev,
+                                                  SpiceMigrateDataSmartcard *mig_data)
 {
     uint8_t *read_data;
 
     spice_debug("read_size  %u", mig_data->read_size);
     read_data = (uint8_t *)mig_data + mig_data->read_data_ptr - sizeof(SpiceMigrateDataHeader);
     if (mig_data->read_size < sizeof(VSCMsgHeader)) {
-        spice_assert(state->buf_size >= mig_data->read_size);
+        spice_assert(dev->priv->buf_size >= mig_data->read_size);
     } else {
-        smartcard_read_buf_prepare(state, (VSCMsgHeader *)read_data);
+        smartcard_read_buf_prepare(dev, (VSCMsgHeader *)read_data);
     }
-    memcpy(state->buf, read_data, mig_data->read_size);
-    state->buf_used = mig_data->read_size;
-    state->buf_pos = state->buf + mig_data->read_size;
+    memcpy(dev->priv->buf, read_data, mig_data->read_size);
+    dev->priv->buf_used = mig_data->read_size;
+    dev->priv->buf_pos = dev->priv->buf + mig_data->read_size;
 }
 
 static int smartcard_channel_client_handle_migrate_data(RedChannelClient *rcc,
@@ -726,7 +671,7 @@ static int smartcard_channel_client_handle_migrate_data(RedChannelClient *rcc,
         return TRUE;
     }
 
-    if (!scc->smartcard_state) {
+    if (!scc->smartcard) {
         SpiceCharDeviceInstance *char_device = smartcard_readers_get_unattached();
 
         if (!char_device) {
@@ -737,10 +682,10 @@ static int smartcard_channel_client_handle_migrate_data(RedChannelClient *rcc,
         }
     }
     spice_debug("reader added %d partial read_size %u", mig_data->reader_added, mig_data->read_size);
-    scc->smartcard_state->reader_added = mig_data->reader_added;
+    scc->smartcard->priv->reader_added = mig_data->reader_added;
 
-    smartcard_device_state_restore_partial_read(scc->smartcard_state, mig_data);
-    return spice_char_device_state_restore(scc->smartcard_state->chardev_st, &mig_data->base);
+    smartcard_device_restore_partial_read(scc->smartcard, mig_data);
+    return red_char_device_restore(RED_CHAR_DEVICE(scc->smartcard), &mig_data->base);
 }
 
 static int smartcard_channel_handle_message(RedChannelClient *rcc,
@@ -793,10 +738,6 @@ static int smartcard_channel_handle_message(RedChannelClient *rcc,
     return TRUE;
 }
 
-static void smartcard_channel_hold_pipe_item(RedChannelClient *rcc, PipeItem *item)
-{
-}
-
 static void smartcard_connect_client(RedChannel *channel, RedClient *client,
                                      RedsStream *stream, int migration,
                                      int num_common_caps, uint32_t *common_caps,
@@ -829,7 +770,7 @@ static void smartcard_connect_client(RedChannel *channel, RedClient *client,
 
 SmartCardChannel *g_smartcard_channel;
 
-static void smartcard_init(void)
+static void smartcard_init(RedsState *reds)
 {
     ChannelCbs channel_cbs = { NULL, };
     ClientCbs client_cbs = { NULL, };
@@ -840,15 +781,15 @@ static void smartcard_init(void)
     channel_cbs.config_socket = smartcard_channel_client_config_socket;
     channel_cbs.on_disconnect = smartcard_channel_on_disconnect;
     channel_cbs.send_item = smartcard_channel_send_item;
-    channel_cbs.hold_item = smartcard_channel_hold_pipe_item;
-    channel_cbs.release_item = smartcard_channel_release_pipe_item;
     channel_cbs.alloc_recv_buf = smartcard_channel_alloc_msg_rcv_buf;
     channel_cbs.release_recv_buf = smartcard_channel_release_msg_rcv_buf;
     channel_cbs.handle_migrate_flush_mark = smartcard_channel_client_handle_migrate_flush_mark;
     channel_cbs.handle_migrate_data = smartcard_channel_client_handle_migrate_data;
 
     g_smartcard_channel = (SmartCardChannel*)red_channel_create(sizeof(SmartCardChannel),
-                                             core, SPICE_CHANNEL_SMARTCARD, 0,
+                                             reds,
+                                             reds_get_core_interface(reds),
+                                             SPICE_CHANNEL_SMARTCARD, 0,
                                              FALSE /* handle_acks */,
                                              smartcard_channel_handle_message,
                                              &channel_cbs,
@@ -859,7 +800,48 @@ static void smartcard_init(void)
     }
 
     client_cbs.connect = smartcard_connect_client;
-    red_channel_register_client_cbs(&g_smartcard_channel->base, &client_cbs);
+    red_channel_register_client_cbs(&g_smartcard_channel->base, &client_cbs, NULL);
 
-    reds_register_channel(&g_smartcard_channel->base);
+    reds_register_channel(reds, &g_smartcard_channel->base);
+}
+
+
+static void
+red_char_device_smartcard_finalize(GObject *object)
+{
+    RedCharDeviceSmartcard *self = RED_CHAR_DEVICE_SMARTCARD(object);
+
+    free(self->priv->buf);
+    if (self->priv->scc) {
+        self->priv->scc->smartcard = NULL;
+    }
+
+    G_OBJECT_CLASS(red_char_device_smartcard_parent_class)->finalize(object);
+}
+
+static void
+red_char_device_smartcard_class_init(RedCharDeviceSmartcardClass *klass)
+{
+    GObjectClass *object_class = G_OBJECT_CLASS(klass);
+    RedCharDeviceClass *char_dev_class = RED_CHAR_DEVICE_CLASS(klass);
+
+    g_type_class_add_private(klass, sizeof (RedCharDeviceSmartcardPrivate));
+
+    object_class->finalize = red_char_device_smartcard_finalize;
+
+    char_dev_class->read_one_msg_from_device = smartcard_read_msg_from_device;
+    char_dev_class->send_msg_to_client = smartcard_send_msg_to_client;
+    char_dev_class->send_tokens_to_client = smartcard_send_tokens_to_client;
+    char_dev_class->remove_client = smartcard_remove_client;
+}
+
+static void
+red_char_device_smartcard_init(RedCharDeviceSmartcard *self)
+{
+    self->priv = RED_CHAR_DEVICE_SMARTCARD_PRIVATE(self);
+
+    self->priv->reader_id = VSCARD_UNDEFINED_READER_ID;
+    self->priv->buf_size = APDUBufSize + sizeof(VSCMsgHeader);
+    self->priv->buf = spice_malloc(self->priv->buf_size);
+    self->priv->buf_pos = self->priv->buf;
 }
