@@ -28,12 +28,15 @@
 #include <netinet/in.h> // IPPROTO_TCP
 #include <netinet/tcp.h> // TCP_NODELAY
 
-#include "common/generated_server_marshallers.h"
+#include <common/generated_server_marshallers.h>
 
-#include "char_device.h"
-#include "red_channel.h"
+#include "char-device.h"
+#include "red-channel.h"
 #include "reds.h"
-#include "migration_protocol.h"
+#include "migration-protocol.h"
+#ifdef USE_LZ4
+#include <lz4.h>
+#endif
 
 /* todo: add flow control. i.e.,
  * (a) limit the tokens available for the client
@@ -41,88 +44,140 @@
  */
 /* 64K should be enough for all but the largest writes + 32 bytes hdr */
 #define BUF_SIZE (64 * 1024 + 32)
+#define COMPRESS_THRESHOLD 1000
 
-typedef struct SpiceVmcPipeItem {
-    PipeItem base;
-    uint32_t refs;
+typedef struct RedVmcPipeItem {
+    RedPipeItem base;
 
+    SpiceDataCompressionType type;
+    uint32_t uncompressed_data_size;
     /* writes which don't fit this will get split, this is not a problem */
     uint8_t buf[BUF_SIZE];
     uint32_t buf_used;
-} SpiceVmcPipeItem;
+} RedVmcPipeItem;
 
 typedef struct SpiceVmcState {
     RedChannel channel; /* Must be the first item */
     RedChannelClient *rcc;
-    SpiceCharDeviceState *chardev_st;
+    RedCharDevice *chardev;
     SpiceCharDeviceInstance *chardev_sin;
-    SpiceVmcPipeItem *pipe_item;
-    SpiceCharDeviceWriteBuffer *recv_from_client_buf;
+    RedVmcPipeItem *pipe_item;
+    RedCharDeviceWriteBuffer *recv_from_client_buf;
     uint8_t port_opened;
 } SpiceVmcState;
 
-typedef struct PortInitPipeItem {
-    PipeItem base;
-    char* name;
-    uint8_t opened;
-} PortInitPipeItem;
+#define RED_TYPE_CHAR_DEVICE_SPICEVMC red_char_device_spicevmc_get_type()
 
-typedef struct PortEventPipeItem {
-    PipeItem base;
-    uint8_t event;
-} PortEventPipeItem;
+#define RED_CHAR_DEVICE_SPICEVMC(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), RED_TYPE_CHAR_DEVICE_SPICEVMC, RedCharDeviceSpiceVmc))
+#define RED_CHAR_DEVICE_SPICEVMC_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), RED_TYPE_CHAR_DEVICE_SPICEVMC, RedCharDeviceSpiceVmcClass))
+#define RED_IS_CHAR_DEVICE_SPICEVMC(obj) (G_TYPE_CHECK_INSTANCE_TYPE((obj), RED_TYPE_CHAR_DEVICE_SPICEVMC))
+#define RED_IS_CHAR_DEVICE_SPICEVMC_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), RED_TYPE_CHAR_DEVICE_SPICEVMC))
+#define RED_CHAR_DEVICE_SPICEVMC_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS((obj), RED_TYPE_CHAR_DEVICE_SPICEVMC, RedCharDeviceSpiceVmcClass))
 
-enum {
-    PIPE_ITEM_TYPE_SPICEVMC_DATA = PIPE_ITEM_TYPE_CHANNEL_BASE,
-    PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA,
-    PIPE_ITEM_TYPE_PORT_INIT,
-    PIPE_ITEM_TYPE_PORT_EVENT,
+typedef struct RedCharDeviceSpiceVmc RedCharDeviceSpiceVmc;
+typedef struct RedCharDeviceSpiceVmcClass RedCharDeviceSpiceVmcClass;
+
+struct RedCharDeviceSpiceVmc {
+    RedCharDevice parent;
 };
 
-static SpiceVmcPipeItem *spicevmc_pipe_item_ref(SpiceVmcPipeItem *item)
+struct RedCharDeviceSpiceVmcClass
 {
-    item->refs++;
-    return item;
-}
+    RedCharDeviceClass parent_class;
+};
 
-static void spicevmc_pipe_item_unref(SpiceVmcPipeItem *item)
+static GType red_char_device_spicevmc_get_type(void) G_GNUC_CONST;
+static RedCharDevice *red_char_device_spicevmc_new(SpiceCharDeviceInstance *sin,
+                                                   RedsState *reds,
+                                                   void *opaque);
+
+G_DEFINE_TYPE(RedCharDeviceSpiceVmc, red_char_device_spicevmc, RED_TYPE_CHAR_DEVICE)
+
+typedef struct RedPortInitPipeItem {
+    RedPipeItem base;
+    char* name;
+    uint8_t opened;
+} RedPortInitPipeItem;
+
+typedef struct RedPortEventPipeItem {
+    RedPipeItem base;
+    uint8_t event;
+} RedPortEventPipeItem;
+
+enum {
+    RED_PIPE_ITEM_TYPE_SPICEVMC_DATA = RED_PIPE_ITEM_TYPE_CHANNEL_BASE,
+    RED_PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA,
+    RED_PIPE_ITEM_TYPE_PORT_INIT,
+    RED_PIPE_ITEM_TYPE_PORT_EVENT,
+};
+
+static void spicevmc_red_channel_release_msg_rcv_buf(RedChannelClient *rcc,
+                                                     uint16_t type,
+                                                     uint32_t size,
+                                                     uint8_t *msg);
+/* n is the data size (uncompressed)
+ * msg_item -- the current pipe item with the uncompressed data
+ * This function returns:
+ *  - NULL upon failure.
+ *  - a new pipe item with the compressed data in it upon success
+ */
+#ifdef USE_LZ4
+static RedVmcPipeItem* try_compress_lz4(SpiceVmcState *state, int n, RedVmcPipeItem *msg_item)
 {
-    if (!--item->refs) {
-        free(item);
+    RedVmcPipeItem *msg_item_compressed;
+    int compressed_data_count;
+
+    if (reds_stream_get_family(state->rcc->stream) == AF_UNIX) {
+        /* AF_LOCAL - data will not be compressed */
+        return NULL;
     }
-}
+    if (n <= COMPRESS_THRESHOLD) {
+        /* n <= threshold - data will not be compressed */
+        return NULL;
+    }
+    if (!red_channel_test_remote_cap(&state->channel, SPICE_SPICEVMC_CAP_DATA_COMPRESS_LZ4)) {
+        /* Client doesn't have compression cap - data will not be compressed */
+        return NULL;
+    }
+    msg_item_compressed = spice_new0(RedVmcPipeItem, 1);
+    red_pipe_item_init(&msg_item_compressed->base, RED_PIPE_ITEM_TYPE_SPICEVMC_DATA);
+    compressed_data_count = LZ4_compress_default((char*)&msg_item->buf,
+                                                 (char*)&msg_item_compressed->buf,
+                                                 n,
+                                                 BUF_SIZE);
 
-SpiceCharDeviceMsgToClient *spicevmc_chardev_ref_msg_to_client(SpiceCharDeviceMsgToClient *msg,
-                                                               void *opaque)
-{
-    return spicevmc_pipe_item_ref((SpiceVmcPipeItem *)msg);
-}
+    if (compressed_data_count > 0 && compressed_data_count < n) {
+        msg_item_compressed->type = SPICE_DATA_COMPRESSION_TYPE_LZ4;
+        msg_item_compressed->uncompressed_data_size = n;
+        msg_item_compressed->buf_used = compressed_data_count;
+        free(msg_item);
+        return msg_item_compressed;
+    }
 
-static void spicevmc_chardev_unref_msg_to_client(SpiceCharDeviceMsgToClient *msg,
-                                                 void *opaque)
-{
-    spicevmc_pipe_item_unref((SpiceVmcPipeItem *)msg);
+    /* LZ4 compression failed or did non compress, fallback a non-compressed data is to be sent */
+    free(msg_item_compressed);
+    return NULL;
 }
+#endif
 
-static SpiceCharDeviceMsgToClient *spicevmc_chardev_read_msg_from_dev(SpiceCharDeviceInstance *sin,
-                                                                      void *opaque)
+static RedPipeItem *spicevmc_chardev_read_msg_from_dev(SpiceCharDeviceInstance *sin,
+                                                       void *opaque)
 {
     SpiceVmcState *state = opaque;
     SpiceCharDeviceInterface *sif;
-    SpiceVmcPipeItem *msg_item;
+    RedVmcPipeItem *msg_item;
     int n;
 
-    sif = SPICE_CONTAINEROF(sin->base.sif, SpiceCharDeviceInterface, base);
+    sif = spice_char_device_get_interface(sin);
 
     if (!state->rcc) {
         return NULL;
     }
 
     if (!state->pipe_item) {
-        msg_item = spice_new0(SpiceVmcPipeItem, 1);
-        msg_item->refs = 1;
-        red_channel_pipe_item_init(&state->channel,
-                                   &msg_item->base, PIPE_ITEM_TYPE_SPICEVMC_DATA);
+        msg_item = spice_new0(RedVmcPipeItem, 1);
+        msg_item->type = SPICE_DATA_COMPRESSION_TYPE_NONE;
+        red_pipe_item_init(&msg_item->base, RED_PIPE_ITEM_TYPE_SPICEVMC_DATA);
     } else {
         spice_assert(state->pipe_item->buf_used == 0);
         msg_item = state->pipe_item;
@@ -133,33 +188,46 @@ static SpiceCharDeviceMsgToClient *spicevmc_chardev_read_msg_from_dev(SpiceCharD
                   sizeof(msg_item->buf));
     if (n > 0) {
         spice_debug("read from dev %d", n);
+#ifdef USE_LZ4
+        RedVmcPipeItem *msg_item_compressed;
+
+        msg_item_compressed = try_compress_lz4(state, n, msg_item);
+        if (msg_item_compressed != NULL) {
+            return &msg_item_compressed->base;
+        }
+#endif
+        msg_item->uncompressed_data_size = n;
         msg_item->buf_used = n;
-        return msg_item;
+        return &msg_item->base;
     } else {
         state->pipe_item = msg_item;
         return NULL;
     }
 }
 
-static void spicevmc_chardev_send_msg_to_client(SpiceCharDeviceMsgToClient *msg,
-                                                 RedClient *client,
-                                                 void *opaque)
+static void spicevmc_chardev_send_msg_to_client(RedPipeItem *msg,
+                                                RedClient *client,
+                                                void *opaque)
 {
     SpiceVmcState *state = opaque;
-    SpiceVmcPipeItem *vmc_msg = msg;
 
     spice_assert(state->rcc->client == client);
-    spicevmc_pipe_item_ref(vmc_msg);
-    red_channel_client_pipe_add_push(state->rcc, &vmc_msg->base);
+    red_pipe_item_ref(msg);
+    red_channel_client_pipe_add_push(state->rcc, msg);
+}
+
+static SpiceVmcState *spicevmc_red_channel_client_get_state(RedChannelClient *rcc)
+{
+    return SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
 }
 
 static void spicevmc_port_send_init(RedChannelClient *rcc)
 {
-    SpiceVmcState *state = SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
+    SpiceVmcState *state = spicevmc_red_channel_client_get_state(rcc);
     SpiceCharDeviceInstance *sin = state->chardev_sin;
-    PortInitPipeItem *item = spice_malloc(sizeof(PortInitPipeItem));
+    RedPortInitPipeItem *item = spice_malloc(sizeof(RedPortInitPipeItem));
 
-    red_channel_pipe_item_init(rcc->channel, &item->base, PIPE_ITEM_TYPE_PORT_INIT);
+    red_pipe_item_init(&item->base, RED_PIPE_ITEM_TYPE_PORT_INIT);
     item->name = strdup(sin->portname);
     item->opened = state->port_opened;
     red_channel_client_pipe_add_push(rcc, &item->base);
@@ -167,9 +235,9 @@ static void spicevmc_port_send_init(RedChannelClient *rcc)
 
 static void spicevmc_port_send_event(RedChannelClient *rcc, uint8_t event)
 {
-    PortEventPipeItem *item = spice_malloc(sizeof(PortEventPipeItem));
+    RedPortEventPipeItem *item = spice_malloc(sizeof(RedPortEventPipeItem));
 
-    red_channel_pipe_item_init(rcc->channel, &item->base, PIPE_ITEM_TYPE_PORT_EVENT);
+    red_pipe_item_init(&item->base, RED_PIPE_ITEM_TYPE_PORT_EVENT);
     item->event = event;
     red_channel_client_pipe_add_push(rcc, &item->base);
 }
@@ -212,23 +280,23 @@ static int spicevmc_red_channel_client_config_socket(RedChannelClient *rcc)
 static void spicevmc_red_channel_client_on_disconnect(RedChannelClient *rcc)
 {
     SpiceVmcState *state;
-    SpiceCharDeviceInstance *sin;
     SpiceCharDeviceInterface *sif;
 
     if (!rcc) {
         return;
     }
 
-    state = SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
-    sin = state->chardev_sin;
-    sif = SPICE_CONTAINEROF(sin->base.sif, SpiceCharDeviceInterface, base);
+    state = spicevmc_red_channel_client_get_state(rcc);
 
-    if (state->chardev_st) {
-        if (spice_char_device_client_exists(state->chardev_st, rcc->client)) {
-            spice_char_device_client_remove(state->chardev_st, rcc->client);
+    /* partial message which wasn't pushed to device */
+    red_char_device_write_buffer_release(state->chardev, &state->recv_from_client_buf);
+
+    if (state->chardev) {
+        if (red_char_device_client_exists(state->chardev, rcc->client)) {
+            red_char_device_client_remove(state->chardev, rcc->client);
         } else {
             spice_printerr("client %p have already been removed from char dev %p",
-                           rcc->client, state->chardev_st);
+                           rcc->client, state->chardev);
         }
     }
 
@@ -238,19 +306,15 @@ static void spicevmc_red_channel_client_on_disconnect(RedChannelClient *rcc)
         red_channel_client_destroy(rcc);
 
     state->rcc = NULL;
+    sif = spice_char_device_get_interface(state->chardev_sin);
     if (sif->state) {
-        sif->state(sin, 0);
+        sif->state(state->chardev_sin, 0);
     }
 }
 
-static SpiceVmcState *spicevmc_red_channel_client_get_state(RedChannelClient *rcc)
-{
-    return SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
-}
-
 static int spicevmc_channel_client_handle_migrate_flush_mark(RedChannelClient *rcc)
 {
-    red_channel_client_pipe_add_type(rcc, PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA);
+    red_channel_client_pipe_add_type(rcc, RED_PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA);
     return TRUE;
 }
 
@@ -273,39 +337,81 @@ static int spicevmc_channel_client_handle_migrate_data(RedChannelClient *rcc,
         spice_error("bad header");
         return FALSE;
     }
-    return spice_char_device_state_restore(state->chardev_st, &mig_data->base);
+    return red_char_device_restore(state->chardev, &mig_data->base);
 }
 
-static int spicevmc_red_channel_client_handle_message(RedChannelClient *rcc,
-                                                      uint16_t type,
-                                                      uint32_t size,
-                                                      uint8_t *msg)
+static int handle_compressed_msg(SpiceVmcState *state, RedChannelClient *rcc,
+                                 SpiceMsgCompressedData *compressed_data_msg)
 {
+    /* NOTE: *decompressed is free by the char-device */
+    int decompressed_size;
+    RedCharDeviceWriteBuffer *write_buf;
+
+    write_buf = red_char_device_write_buffer_get(state->chardev, rcc->client,
+                                                 compressed_data_msg->uncompressed_size);
+    if (!write_buf) {
+        return FALSE;
+    }
+
+    switch (compressed_data_msg->type) {
+#ifdef USE_LZ4
+    case SPICE_DATA_COMPRESSION_TYPE_LZ4: {
+        uint8_t *decompressed = write_buf->buf;
+        decompressed_size = LZ4_decompress_safe ((char *)compressed_data_msg->compressed_data,
+                                                 (char *)decompressed,
+                                                 compressed_data_msg->compressed_size,
+                                                 compressed_data_msg->uncompressed_size);
+        break;
+    }
+#endif
+    default:
+        spice_warning("Invalid Compression Type");
+        red_char_device_write_buffer_release(state->chardev, &write_buf);
+        return FALSE;
+    }
+    if (decompressed_size != compressed_data_msg->uncompressed_size) {
+        spice_warning("Decompression Error");
+        red_char_device_write_buffer_release(state->chardev, &write_buf);
+        return FALSE;
+    }
+    write_buf->buf_used = decompressed_size;
+    red_char_device_write_buffer_add(state->chardev, write_buf);
+    return TRUE;
+}
+
+static int spicevmc_red_channel_client_handle_message_parsed(RedChannelClient *rcc,
+                                                             uint32_t size,
+                                                             uint16_t type,
+                                                             void *msg)
+{
+    /* NOTE: *msg free by free() (when cb to spicevmc_red_channel_release_msg_rcv_buf
+     * with the compressed msg type) */
     SpiceVmcState *state;
-    SpiceCharDeviceInstance *sin;
     SpiceCharDeviceInterface *sif;
 
     state = spicevmc_red_channel_client_get_state(rcc);
-    sin = state->chardev_sin;
-    sif = SPICE_CONTAINEROF(sin->base.sif, SpiceCharDeviceInterface, base);
+    sif = spice_char_device_get_interface(state->chardev_sin);
 
     switch (type) {
     case SPICE_MSGC_SPICEVMC_DATA:
         spice_assert(state->recv_from_client_buf->buf == msg);
         state->recv_from_client_buf->buf_used = size;
-        spice_char_device_write_buffer_add(state->chardev_st, state->recv_from_client_buf);
+        red_char_device_write_buffer_add(state->chardev, state->recv_from_client_buf);
         state->recv_from_client_buf = NULL;
         break;
+    case SPICE_MSGC_SPICEVMC_COMPRESSED_DATA:
+        return handle_compressed_msg(state, rcc, (SpiceMsgCompressedData*)msg);
+        break;
     case SPICE_MSGC_PORT_EVENT:
         if (size != sizeof(uint8_t)) {
             spice_warning("bad port event message size");
             return FALSE;
         }
         if (sif->base.minor_version >= 2 && sif->event != NULL)
-            sif->event(sin, *msg);
+            sif->event(state->chardev_sin, *(uint8_t*)msg);
         break;
     default:
-        return red_channel_client_handle_message(rcc, size, type, msg);
+        return red_channel_client_handle_message(rcc, size, type, (uint8_t*)msg);
     }
 
     return TRUE;
@@ -317,15 +423,15 @@ static uint8_t *spicevmc_red_channel_alloc_msg_rcv_buf(RedChannelClient *rcc,
 {
     SpiceVmcState *state;
 
-    state = SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
+    state = spicevmc_red_channel_client_get_state(rcc);
 
     switch (type) {
     case SPICE_MSGC_SPICEVMC_DATA:
         assert(!state->recv_from_client_buf);
 
-        state->recv_from_client_buf = spice_char_device_write_buffer_get(state->chardev_st,
-                                                                         rcc->client,
-                                                                         size);
+        state->recv_from_client_buf = red_char_device_write_buffer_get(state->chardev,
+                                                                       rcc->client,
+                                                                       size);
         if (!state->recv_from_client_buf) {
             spice_error("failed to allocate write buffer");
             return NULL;
@@ -345,55 +451,58 @@ static void spicevmc_red_channel_release_msg_rcv_buf(RedChannelClient *rcc,
 {
     SpiceVmcState *state;
 
-    state = SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
+    state = spicevmc_red_channel_client_get_state(rcc);
 
     switch (type) {
     case SPICE_MSGC_SPICEVMC_DATA:
-        if (state->recv_from_client_buf) { /* buffer wasn't pushed to device */
-            spice_char_device_write_buffer_release(state->chardev_st, state->recv_from_client_buf);
-            state->recv_from_client_buf = NULL;
-        }
+        /* buffer wasn't pushed to device */
+        red_char_device_write_buffer_release(state->chardev, &state->recv_from_client_buf);
         break;
     default:
         free(msg);
     }
 }
 
-static void spicevmc_red_channel_hold_pipe_item(RedChannelClient *rcc,
-    PipeItem *item)
-{
-    /* NOOP */
-}
-
 static void spicevmc_red_channel_send_data(RedChannelClient *rcc,
                                            SpiceMarshaller *m,
-                                           PipeItem *item)
+                                           RedPipeItem *item)
 {
-    SpiceVmcPipeItem *i = SPICE_CONTAINEROF(item, SpiceVmcPipeItem, base);
+    RedVmcPipeItem *i = SPICE_UPCAST(RedVmcPipeItem, item);
 
-    red_channel_client_init_send_data(rcc, SPICE_MSG_SPICEVMC_DATA, item);
+    /* for compatibility send using not compressed data message */
+    if (i->type == SPICE_DATA_COMPRESSION_TYPE_NONE) {
+        red_channel_client_init_send_data(rcc, SPICE_MSG_SPICEVMC_DATA, item);
+    } else {
+        /* send as compressed */
+        red_channel_client_init_send_data(rcc, SPICE_MSG_SPICEVMC_COMPRESSED_DATA, item);
+        SpiceMsgCompressedData compressed_msg = {
+            .type = i->type,
+            .uncompressed_size = i->uncompressed_data_size
+        };
+        spice_marshall_SpiceMsgCompressedData(m, &compressed_msg);
+    }
     spice_marshaller_add_ref(m, i->buf, i->buf_used);
 }
 
 static void spicevmc_red_channel_send_migrate_data(RedChannelClient *rcc,
                                                    SpiceMarshaller *m,
-                                                   PipeItem *item)
+                                                   RedPipeItem *item)
 {
     SpiceVmcState *state;
 
-    state = SPICE_CONTAINEROF(rcc->channel, SpiceVmcState, channel);
+    state = spicevmc_red_channel_client_get_state(rcc);
     red_channel_client_init_send_data(rcc, SPICE_MSG_MIGRATE_DATA, item);
     spice_marshaller_add_uint32(m, SPICE_MIGRATE_DATA_SPICEVMC_MAGIC);
     spice_marshaller_add_uint32(m, SPICE_MIGRATE_DATA_SPICEVMC_VERSION);
 
-    spice_char_device_state_migrate_data_marshall(state->chardev_st, m);
+    red_char_device_migrate_data_marshall(state->chardev, m);
 }
 
 static void spicevmc_red_channel_send_port_init(RedChannelClient *rcc,
                                                 SpiceMarshaller *m,
-                                                PipeItem *item)
+                                                RedPipeItem *item)
 {
-    PortInitPipeItem *i = SPICE_CONTAINEROF(item, PortInitPipeItem, base);
+    RedPortInitPipeItem *i = SPICE_UPCAST(RedPortInitPipeItem, item);
     SpiceMsgPortInit init;
 
     red_channel_client_init_send_data(rcc, SPICE_MSG_PORT_INIT, item);
@@ -405,9 +514,9 @@ static void spicevmc_red_channel_send_port_init(RedChannelClient *rcc,
 
 static void spicevmc_red_channel_send_port_event(RedChannelClient *rcc,
                                                  SpiceMarshaller *m,
-                                                 PipeItem *item)
+                                                 RedPipeItem *item)
 {
-    PortEventPipeItem *i = SPICE_CONTAINEROF(item, PortEventPipeItem, base);
+    RedPortEventPipeItem *i = SPICE_UPCAST(RedPortEventPipeItem, item);
     SpiceMsgPortEvent event;
 
     red_channel_client_init_send_data(rcc, SPICE_MSG_PORT_EVENT, item);
@@ -416,41 +525,30 @@ static void spicevmc_red_channel_send_port_event(RedChannelClient *rcc,
 }
 
 static void spicevmc_red_channel_send_item(RedChannelClient *rcc,
-                                           PipeItem *item)
+                                           RedPipeItem *item)
 {
     SpiceMarshaller *m = red_channel_client_get_marshaller(rcc);
 
     switch (item->type) {
-    case PIPE_ITEM_TYPE_SPICEVMC_DATA:
+    case RED_PIPE_ITEM_TYPE_SPICEVMC_DATA:
         spicevmc_red_channel_send_data(rcc, m, item);
         break;
-    case PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA:
+    case RED_PIPE_ITEM_TYPE_SPICEVMC_MIGRATE_DATA:
         spicevmc_red_channel_send_migrate_data(rcc, m, item);
         break;
-    case PIPE_ITEM_TYPE_PORT_INIT:
+    case RED_PIPE_ITEM_TYPE_PORT_INIT:
         spicevmc_red_channel_send_port_init(rcc, m, item);
         break;
-    case PIPE_ITEM_TYPE_PORT_EVENT:
+    case RED_PIPE_ITEM_TYPE_PORT_EVENT:
         spicevmc_red_channel_send_port_event(rcc, m, item);
         break;
     default:
         spice_error("bad pipe item %d", item->type);
-        free(item);
         return;
     }
     red_channel_client_begin_send_message(rcc);
 }
 
-static void spicevmc_red_channel_release_pipe_item(RedChannelClient *rcc,
-    PipeItem *item, int item_pushed)
-{
-    if (item->type == PIPE_ITEM_TYPE_SPICEVMC_DATA) {
-        spicevmc_pipe_item_unref((SpiceVmcPipeItem *)item);
-    } else {
-        free(item);
-    }
-}
-
 static void spicevmc_connect(RedChannel *channel, RedClient *client,
     RedsStream *stream, int migration, int num_common_caps,
     uint32_t *common_caps, int num_caps, uint32_t *caps)
@@ -462,7 +560,6 @@ static void spicevmc_connect(RedChannel *channel, RedClient *client,
 
     state = SPICE_CONTAINEROF(channel, SpiceVmcState, channel);
     sin = state->chardev_sin;
-    sif = SPICE_CONTAINEROF(sin->base.sif, SpiceCharDeviceInterface, base);
 
     if (state->rcc) {
         spice_printerr("channel client %d:%d (%p) already connected, refusing second connection",
@@ -487,80 +584,74 @@ static void spicevmc_connect(RedChannel *channel, RedClient *client,
         spicevmc_port_send_init(rcc);
     }
 
-    if (!spice_char_device_client_add(state->chardev_st, client, FALSE, 0, ~0, ~0,
-                                      red_channel_client_waits_for_migrate_data(rcc))) {
+    if (!red_char_device_client_add(state->chardev, client, FALSE, 0, ~0, ~0,
+                                    red_channel_client_is_waiting_for_migrate_data(rcc))) {
         spice_warning("failed to add client to spicevmc");
         red_channel_client_disconnect(rcc);
         return;
     }
 
+    sif = spice_char_device_get_interface(state->chardev_sin);
     if (sif->state) {
         sif->state(sin, 1);
     }
 }
 
-SpiceCharDeviceState *spicevmc_device_connect(SpiceCharDeviceInstance *sin,
-                                              uint8_t channel_type)
+RedCharDevice *spicevmc_device_connect(RedsState *reds,
+                                       SpiceCharDeviceInstance *sin,
+                                       uint8_t channel_type)
 {
     static uint8_t id[256] = { 0, };
     SpiceVmcState *state;
     ChannelCbs channel_cbs = { NULL, };
     ClientCbs client_cbs = { NULL, };
-    SpiceCharDeviceCallbacks char_dev_cbs = {NULL, };
 
     channel_cbs.config_socket = spicevmc_red_channel_client_config_socket;
     channel_cbs.on_disconnect = spicevmc_red_channel_client_on_disconnect;
     channel_cbs.send_item = spicevmc_red_channel_send_item;
-    channel_cbs.hold_item = spicevmc_red_channel_hold_pipe_item;
-    channel_cbs.release_item = spicevmc_red_channel_release_pipe_item;
     channel_cbs.alloc_recv_buf = spicevmc_red_channel_alloc_msg_rcv_buf;
     channel_cbs.release_recv_buf = spicevmc_red_channel_release_msg_rcv_buf;
     channel_cbs.handle_migrate_flush_mark = spicevmc_channel_client_handle_migrate_flush_mark;
     channel_cbs.handle_migrate_data = spicevmc_channel_client_handle_migrate_data;
 
-    state = (SpiceVmcState*)red_channel_create(sizeof(SpiceVmcState),
-                                   core, channel_type, id[channel_type]++,
+    state = (SpiceVmcState*)red_channel_create_parser(sizeof(SpiceVmcState), reds,
+                                   reds_get_core_interface(reds), channel_type, id[channel_type]++,
                                    FALSE /* handle_acks */,
-                                   spicevmc_red_channel_client_handle_message,
+                                   spice_get_client_channel_parser(channel_type, NULL),
+                                   spicevmc_red_channel_client_handle_message_parsed,
                                    &channel_cbs,
                                    SPICE_MIGRATE_NEED_FLUSH | SPICE_MIGRATE_NEED_DATA_TRANSFER);
     red_channel_init_outgoing_messages_window(&state->channel);
 
     client_cbs.connect = spicevmc_connect;
-    red_channel_register_client_cbs(&state->channel, &client_cbs);
+    red_channel_register_client_cbs(&state->channel, &client_cbs, NULL);
+#ifdef USE_LZ4
+    red_channel_set_cap(&state->channel, SPICE_SPICEVMC_CAP_DATA_COMPRESS_LZ4);
+#endif
 
-    char_dev_cbs.read_one_msg_from_device = spicevmc_chardev_read_msg_from_dev;
-    char_dev_cbs.ref_msg_to_client = spicevmc_chardev_ref_msg_to_client;
-    char_dev_cbs.unref_msg_to_client = spicevmc_chardev_unref_msg_to_client;
-    char_dev_cbs.send_msg_to_client = spicevmc_chardev_send_msg_to_client;
-    char_dev_cbs.send_tokens_to_client = spicevmc_char_dev_send_tokens_to_client;
-    char_dev_cbs.remove_client = spicevmc_char_dev_remove_client;
-
-    state->chardev_st = spice_char_device_state_create(sin,
-                                                       0, /* tokens interval */
-                                                       ~0, /* self tokens */
-                                                       &char_dev_cbs,
-                                                       state);
+    state->chardev = red_char_device_spicevmc_new(sin, reds, state);
     state->chardev_sin = sin;
 
-    reds_register_channel(&state->channel);
-    return state->chardev_st;
+    reds_register_channel(reds, &state->channel);
+    return state->chardev;
 }
 
 /* Must be called from RedClient handling thread. */
-void spicevmc_device_disconnect(SpiceCharDeviceInstance *sin)
+void spicevmc_device_disconnect(RedsState *reds, SpiceCharDeviceInstance *sin)
 {
     SpiceVmcState *state;
 
-    state = (SpiceVmcState *)spice_char_device_state_opaque_get(sin->st);
+    /* FIXME */
+    state = (SpiceVmcState *)red_char_device_opaque_get((RedCharDevice*)sin->st);
 
-    if (state->recv_from_client_buf) {
-        spice_char_device_write_buffer_release(state->chardev_st, state->recv_from_client_buf);
-    }
-    spice_char_device_state_destroy(sin->st);
-    state->chardev_st = NULL;
+    red_char_device_write_buffer_release(state->chardev, &state->recv_from_client_buf);
 
-    reds_unregister_channel(&state->channel);
+    /* FIXME */
+    red_char_device_destroy((RedCharDevice*)sin->st);
+    state->chardev = NULL;
+    sin->st = NULL;
+
+    reds_unregister_channel(reds, &state->channel);
     free(state->pipe_item);
     red_channel_destroy(&state->channel);
 }
@@ -569,7 +660,13 @@ SPICE_GNUC_VISIBLE void spice_server_port_event(SpiceCharDeviceInstance *sin, ui
 {
     SpiceVmcState *state;
 
-    state = (SpiceVmcState *)spice_char_device_state_opaque_get(sin->st);
+    if (sin->st == NULL) {
+        spice_warning("no SpiceCharDeviceState attached to instance %p", sin);
+        return;
+    }
+
+    /* FIXME */
+    state = (SpiceVmcState *)red_char_device_opaque_get((RedCharDevice*)sin->st);
     if (event == SPICE_PORT_EVENT_OPENED) {
         state->port_opened = TRUE;
     } else if (event == SPICE_PORT_EVENT_CLOSED) {
@@ -582,3 +679,33 @@ SPICE_GNUC_VISIBLE void spice_server_port_event(SpiceCharDeviceInstance *sin, ui
 
     spicevmc_port_send_event(state->rcc, event);
 }
+
+static void
+red_char_device_spicevmc_class_init(RedCharDeviceSpiceVmcClass *klass)
+{
+    RedCharDeviceClass *char_dev_class = RED_CHAR_DEVICE_CLASS(klass);
+
+    char_dev_class->read_one_msg_from_device = spicevmc_chardev_read_msg_from_dev;
+    char_dev_class->send_msg_to_client = spicevmc_chardev_send_msg_to_client;
+    char_dev_class->send_tokens_to_client = spicevmc_char_dev_send_tokens_to_client;
+    char_dev_class->remove_client = spicevmc_char_dev_remove_client;
+}
+
+static void
+red_char_device_spicevmc_init(RedCharDeviceSpiceVmc *self)
+{
+}
+
+static RedCharDevice *
+red_char_device_spicevmc_new(SpiceCharDeviceInstance *sin,
+                             RedsState *reds,
+                             void *opaque)
+{
+    return g_object_new(RED_TYPE_CHAR_DEVICE_SPICEVMC,
+                        "sin", sin,
+                        "spice-server", reds,
+                        "client-tokens-interval", 0ULL,
+                        "self-tokens", ~0ULL,
+                        "opaque", opaque,
+                        NULL);
+}
