@@ -23,6 +23,8 @@
 #include <config.h>
 #endif
 
+#include <glib.h>
+
 #include "quic.h"
 #include "spice_common.h"
 #include "bitops.h"
@@ -35,7 +37,8 @@
 //#define RLE_PRED_3
 #define QUIC_RGB
 
-#define QUIC_MAGIC (*(uint32_t *)"QUIC")
+/* ASCII "QUIC" */
+#define QUIC_MAGIC 0x43495551
 #define QUIC_VERSION_MAJOR 0U
 #define QUIC_VERSION_MINOR 1U
 #define QUIC_VERSION ((QUIC_VERSION_MAJOR << 16) | (QUIC_VERSION_MAJOR & 0xffff))
@@ -75,6 +78,9 @@ typedef struct QuicFamily {
     unsigned int notGRsuffixlen[MAXNUMCODES];    /* indexed by code number, contains suffix
                                                     length of the not-GR codeword */
 
+    unsigned int golomb_code_len[256][MAXNUMCODES];
+    unsigned int golomb_code[256][MAXNUMCODES];
+
     /* array for translating distribution U to L for depths up to 8 bpp,
     initialized by decorelateinit() */
     BYTE xlatU2L[256];
@@ -188,16 +194,7 @@ static const unsigned long int bppmask[33] = {
     0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff /* [32] */
 };
 
-static const unsigned int bitat[32] = {
-    0x00000001, 0x00000002, 0x00000004, 0x00000008,
-    0x00000010, 0x00000020, 0x00000040, 0x00000080,
-    0x00000100, 0x00000200, 0x00000400, 0x00000800,
-    0x00001000, 0x00002000, 0x00004000, 0x00008000,
-    0x00010000, 0x00020000, 0x00040000, 0x00080000,
-    0x00100000, 0x00200000, 0x00400000, 0x00800000,
-    0x01000000, 0x02000000, 0x04000000, 0x08000000,
-    0x10000000, 0x20000000, 0x40000000, 0x80000000 /* [31]*/
-};
+#define bitat(n) (1u<<(n))
 
 
 #define TABRAND_TABSIZE 256
@@ -360,9 +357,22 @@ static void corelate_init(QuicFamily *family, int bpc)
     }
 }
 
+static void golomb_coding_slow(QuicFamily *family, const BYTE n, const unsigned int l,
+                               unsigned int * const codeword,
+                               unsigned int * const codewordlen)
+{
+    if (n < family->nGRcodewords[l]) {
+        (*codeword) = bitat(l) | (n & bppmask[l]);
+        (*codewordlen) = (n >> l) + l + 1;
+    } else {
+        (*codeword) = n - family->nGRcodewords[l];
+        (*codewordlen) = family->notGRcwlen[l];
+    }
+}
+
 static void family_init(QuicFamily *family, int bpc, int limit)
 {
-    int l;
+    int l, b;
 
     for (l = 0; l < bpc; l++) { /* fill arrays indexed by code number */
         int altprefixlen, altcodewords;
@@ -375,9 +385,16 @@ static void family_init(QuicFamily *family, int bpc, int limit)
         altcodewords = bppmask[bpc] + 1 - (altprefixlen << l);
 
         family->nGRcodewords[l] = (altprefixlen << l);
-        family->notGRcwlen[l] = altprefixlen + ceil_log_2(altcodewords);
-        family->notGRprefixmask[l] = bppmask[32 - altprefixlen]; /* needed for decoding only */
         family->notGRsuffixlen[l] = ceil_log_2(altcodewords); /* needed for decoding only */
+        family->notGRcwlen[l] = altprefixlen + family->notGRsuffixlen[l];
+        family->notGRprefixmask[l] = bppmask[32 - altprefixlen]; /* needed for decoding only */
+
+        for (b = 0; b < 256; b++) {
+            unsigned int code, len;
+            golomb_coding_slow(family, b, l, &code, &len);
+            family->golomb_code[b][l] = code;
+            family->golomb_code_len[b][l] = len;
+        }
     }
 
     decorelate_init(family, bpc);
@@ -405,7 +422,7 @@ static void __write_io_word(Encoder *encoder)
 
 static void (*__write_io_word_ptr)(Encoder *encoder) = __write_io_word;
 
-static INLINE void write_io_word(Encoder *encoder)
+static inline void write_io_word(Encoder *encoder)
 {
     if (encoder->io_now == encoder->io_end) {
         __write_io_word_ptr(encoder); //disable inline optimizations
@@ -414,7 +431,7 @@ static INLINE void write_io_word(Encoder *encoder)
     *(encoder->io_now++) = encoder->io_word;
 }
 
-static INLINE void encode(Encoder *encoder, unsigned int word, unsigned int len)
+static inline void encode(Encoder *encoder, unsigned int word, unsigned int len)
 {
     int delta;
 
@@ -435,13 +452,13 @@ static INLINE void encode(Encoder *encoder, unsigned int word, unsigned int len)
     spice_assert((encoder->io_word & bppmask[encoder->io_available_bits]) == 0);
 }
 
-static INLINE void encode_32(Encoder *encoder, unsigned int word)
+static inline void encode_32(Encoder *encoder, unsigned int word)
 {
     encode(encoder, word >> 16, 16);
     encode(encoder, word & 0x0000ffff, 16);
 }
 
-static INLINE void flush(Encoder *encoder)
+static inline void flush(Encoder *encoder)
 {
     if (encoder->io_available_bits > 0 && encoder->io_available_bits != 32) {
         encode(encoder, 0, encoder->io_available_bits);
@@ -453,23 +470,23 @@ static INLINE void flush(Encoder *encoder)
 static void __read_io_word(Encoder *encoder)
 {
     more_io_words(encoder);
-    encoder->io_next_word = *(encoder->io_now++);
+    encoder->io_next_word = GUINT32_FROM_LE(*(encoder->io_now++));
 }
 
 static void (*__read_io_word_ptr)(Encoder *encoder) = __read_io_word;
 
 
-static INLINE void read_io_word(Encoder *encoder)
+static inline void read_io_word(Encoder *encoder)
 {
     if (encoder->io_now == encoder->io_end) {
         __read_io_word_ptr(encoder); //disable inline optimizations
         return;
     }
     spice_assert(encoder->io_now < encoder->io_end);
-    encoder->io_next_word = *(encoder->io_now++);
+    encoder->io_next_word = GUINT32_FROM_LE(*(encoder->io_now++));
 }
 
-static INLINE void decode_eatbits(Encoder *encoder, int len)
+static inline void decode_eatbits(Encoder *encoder, int len)
 {
     int delta;
 
@@ -489,7 +506,7 @@ static INLINE void decode_eatbits(Encoder *encoder, int len)
     encoder->io_word |= (encoder->io_next_word >> encoder->io_available_bits);
 }
 
-static INLINE void decode_eat32bits(Encoder *encoder)
+static inline void decode_eat32bits(Encoder *encoder)
 {
     decode_eatbits(encoder, 16);
     decode_eatbits(encoder, 16);
@@ -499,7 +516,7 @@ static INLINE void decode_eat32bits(Encoder *encoder)
 
 #ifdef RLE_STAT
 
-static INLINE void encode_ones(Encoder *encoder, unsigned int n)
+static inline void encode_ones(Encoder *encoder, unsigned int n)
 {
     unsigned int count;
 
@@ -514,31 +531,11 @@ static INLINE void encode_ones(Encoder *encoder, unsigned int n)
 
 #define MELCSTATES 32 /* number of melcode states */
 
-static int zeroLUT[256]; /* table to find out number of leading zeros */
-
-static int J[MELCSTATES] = {
+static const int J[MELCSTATES] = {
     0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 5, 5, 6, 6, 7,
     7, 8, 9, 10, 11, 12, 13, 14, 15
 };
 
-/* creates the bit counting look-up table. */
-static void init_zeroLUT(void)
-{
-    int i, j, k, l;
-
-    j = k = 1;
-    l = 8;
-    for (i = 0; i < 256; ++i) {
-        zeroLUT[i] = l;
-        --k;
-        if (k == 0) {
-            k = j;
-            --l;
-            j *= 2;
-        }
-    }
-}
-
 static void encoder_init_rle(CommonState *state)
 {
     state->melcstate = 0;
@@ -555,7 +552,7 @@ static void encode_run(Encoder *encoder, unsigned int runlen) //todo: try use en
     while (runlen >= encoder->rgb_state.melcorder) {
         hits++;
         runlen -= encoder->rgb_state.melcorder;
-        if (encoder->rgb_state.melcstate < MELCSTATES) {
+        if (encoder->rgb_state.melcstate < MELCSTATES - 1) {
             encoder->rgb_state.melclen = J[++encoder->rgb_state.melcstate];
             encoder->rgb_state.melcorder = (1L << encoder->rgb_state.melclen);
         }
@@ -587,7 +584,7 @@ static void encode_channel_run(Encoder *encoder, Channel *channel, unsigned int
     while (runlen >= channel->state.melcorder) {
         hits++;
         runlen -= channel->state.melcorder;
-        if (channel->state.melcstate < MELCSTATES) {
+        if (channel->state.melcstate < MELCSTATES - 1) {
             channel->state.melclen = J[++channel->state.melcstate];
             channel->state.melcorder = (1L << channel->state.melclen);
         }
@@ -619,12 +616,12 @@ static int decode_run(Encoder *encoder)
 
     do {
         register int temp, hits;
-        temp = zeroLUT[(BYTE)(~(encoder->io_word >> 24))];/* number of leading ones in the
+        temp = lzeroes[(BYTE)(~(encoder->io_word >> 24))];/* number of leading ones in the
                                                                       input stream, up to 8 */
         for (hits = 1; hits <= temp; hits++) {
             runlen += encoder->rgb_state.melcorder;
 
-            if (encoder->rgb_state.melcstate < MELCSTATES) {
+            if (encoder->rgb_state.melcstate < MELCSTATES - 1) {
                 encoder->rgb_state.melclen = J[++encoder->rgb_state.melcstate];
                 encoder->rgb_state.melcorder = (1U << encoder->rgb_state.melclen);
             }
@@ -660,12 +657,12 @@ static int decode_channel_run(Encoder *encoder, Channel *channel)
 
     do {
         register int temp, hits;
-        temp = zeroLUT[(BYTE)(~(encoder->io_word >> 24))];/* number of leading ones in the
+        temp = lzeroes[(BYTE)(~(encoder->io_word >> 24))];/* number of leading ones in the
                                                                       input stream, up to 8 */
         for (hits = 1; hits <= temp; hits++) {
             runlen += channel->state.melcorder;
 
-            if (channel->state.melcstate < MELCSTATES) {
+            if (channel->state.melcstate < MELCSTATES - 1) {
                 channel->state.melclen = J[++channel->state.melcstate];
                 channel->state.melcorder = (1U << channel->state.melclen);
             }
@@ -695,7 +692,7 @@ static int decode_channel_run(Encoder *encoder, Channel *channel)
 
 #else
 
-static INLINE void encode_run(Encoder *encoder, unsigned int len)
+static inline void encode_run(Encoder *encoder, unsigned int len)
 {
     int odd = len & 1U;
     int msb;
@@ -716,7 +713,7 @@ static INLINE void encode_run(Encoder *encoder, unsigned int len)
     }
 }
 
-static INLINE unsigned int decode_run(Encoder *encoder)
+static inline unsigned int decode_run(Encoder *encoder)
 {
     unsigned int len = 0;
     int count;
@@ -738,9 +735,9 @@ static INLINE unsigned int decode_run(Encoder *encoder)
 #endif
 #endif
 
-static INLINE void init_decode_io(Encoder *encoder)
+static inline void init_decode_io(Encoder *encoder)
 {
-    encoder->io_next_word = encoder->io_word = *(encoder->io_now++);
+    encoder->io_next_word = encoder->io_word = GUINT32_FROM_LE(*(encoder->io_now++));
     encoder->io_available_bits = 0;
 }
 
@@ -817,7 +814,7 @@ typedef uint16_t rgb16_pixel_t;
 
 #endif
 
-static void fill_model_structures(Encoder *encoder, FamilyStat *family_stat,
+static void fill_model_structures(SPICE_GNUC_UNUSED Encoder *encoder, FamilyStat *family_stat,
                                   unsigned int rep_first, unsigned int first_size,
                                   unsigned int rep_next, unsigned int mul_size,
                                   unsigned int levels, unsigned int ncounters,
@@ -888,9 +885,12 @@ static void find_model_params(Encoder *encoder,
     unsigned int bstart, bend = 0;   /* bucket start and end, range : 0 to levels-1*/
     unsigned int repcntr;            /* helper */
 
+    /* The only valid values are 1, 3 and 5.
+       0, 2 and 4 are obsolete and the rest of the
+       values are considered out of the range. */
+    spice_static_assert (evol == 1 || evol == 3 || evol == 5);
     spice_assert(bpc <= 8 && bpc > 0);
 
-
     *ncounters = 8;
 
     *levels = 0x1 << bpc;
@@ -916,12 +916,9 @@ static void find_model_params(Encoder *encoder,
         *repnext = 1;
         *mulsize = 4;
         break;
-    case 0: /* obsolete */
-    case 2: /* obsolete */
-    case 4: /* obsolete */
-        encoder->usr->error(encoder->usr, "findmodelparams(): evol value obsolete!!!\n");
     default:
         encoder->usr->error(encoder->usr, "findmodelparams(): evol out of range!!!\n");
+        return;
     }
 
     *nbuckets = 0;
@@ -929,7 +926,7 @@ static void find_model_params(Encoder *encoder,
     bsize = *firstsize;
 
     do { /* other buckets */
-        if (nbuckets) {         /* bucket start */
+        if (*nbuckets) {         /* bucket start */
             bstart = bend + 1;
         } else {
             bstart = 0;
@@ -1076,7 +1073,7 @@ static int init_encoder(Encoder *encoder, QuicUsrContext *usr)
 
 static int encoder_reste(Encoder *encoder, uint32_t *io_ptr, uint32_t *io_ptr_end)
 {
-    spice_assert(((unsigned long)io_ptr % 4) == ((unsigned long)io_ptr_end % 4));
+    spice_assert(((uintptr_t)io_ptr % 4) == ((uintptr_t)io_ptr_end % 4));
     spice_assert(io_ptr <= io_ptr_end);
 
     encoder->rgb_state.waitcnt = 0;
@@ -1637,13 +1634,11 @@ int quic_decode(QuicContext *quic, QuicImageType type, uint8_t *buf, int stride)
     return QUIC_OK;
 }
 
-static int need_init = TRUE;
-
 QuicContext *quic_create(QuicUsrContext *usr)
 {
     Encoder *encoder;
 
-    if (!usr || need_init || !usr->error || !usr->warn || !usr->info || !usr->malloc ||
+    if (!usr || !usr->error || !usr->warn || !usr->info || !usr->malloc ||
         !usr->free || !usr->more_space || !usr->more_lines) {
         return NULL;
     }
@@ -1674,16 +1669,8 @@ void quic_destroy(QuicContext *quic)
     encoder->usr->free(encoder->usr, encoder);
 }
 
-void quic_init(void)
+SPICE_CONSTRUCTOR_FUNC(quic_global_init)
 {
-    if (!need_init) {
-        return;
-    }
-    need_init = FALSE;
-
     family_init(&family_8bpc, 8, DEFmaxclen);
     family_init(&family_5bpc, 5, DEFmaxclen);
-#if defined(RLE) && defined(RLE_STAT)
-    init_zeroLUT();
-#endif
 }
